/* Copyright 2011, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS 
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, 
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, 
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL, 
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */

/* Core for unique ADC configuration present first on a TS-7558.  A single
 * high speed 200ksps 12-bit SPI ADC is connected to a 16 channel analog
 * mux.  After the mux, there is an op-amp circuit with a FET connected
 * under FPGA control that adds 2x gain stage before feeding to the 
 * SPI National Semiconductor ADC121S021 ADC chip.
 *
 * This core takes continuous samples of selected channels and logs them
 * to an external 16-bit 8kbyte WISHBONE memory slave at a programmable
 * clock rate up to 20Khz.  Each 12-bit sample is logged to external memory
 * in the following format:
 *
 *   bits 15-12: ADC channel (0-15)
 *   bits 11-0: 12-bit ADC reading
 *
 * Upon reaching the end of the 8Kbyte buffer, the address is wrapped back
 * to zero and sampling continues.  Software must be able to keep up and
 * read samples before the address wraps around and old samples are
 * overwritten.  There will be an irq every 4kbytes written to the external
 * memory.  Any IRQ is cleared and re-armed upon reading of the current
 * state reg.
 *
 * Register map:
 * base + 0x0: channel enable #15 (MSB) through #0 (LSB) (1- enabled) (RW)
 * base + 0x2: current state  
 *   bit 15: sampling enable (RW)
 *   bit 14: overflow (RO - self clears after read)
 *   bit 13: interrupt mood (1 - interested, 0 - not-interested) (RW)
 *   bit 12: reserved (write as 0)
 *   bits 11-0: next buffer num to fill (0-4095 for 8kbyte total buffer) (RO)
 * base + 0x4: sample period (RW)
 *   bit 15: timescale (1 - ms, 0 - us) (RW)
 *   bits 14-0: value (4096 is 4097us period) (RW)
 * base + 0x6: misc
 *   bit 15: external sample trigger enable
 *   bits 14-12: external trigger input num (0-7) (trigs on posedge)
 *   bits 11-8: current loop enable on chans #3 (MSB) - #0 (LSB) (RW)
 *   bits 7-0: gain 2x enable on chans #7 (MSB) to #0 (LSB) (RW)
 * base + 0x8: interest criteria, CPU is interested ... (RW)
 *   bit 15: ... when any digital input changes 
 *   bit 14: ... when any enabled counters change
 *   bit 13: ... when a quadrature counter changes direction
 *   bit 12: ... when a quadrature counter changes value
 *   bit 11: ... when one particular digital input ... 
 *   bits 10-9: 
 *     0 - ... changes state (change from last sample to cur sample)
 *     1 - ... glitches (a change >= 25ns but <= sample period)
 *     2 - ... is high
 *     3 - ... is low
 *   bits 8-5: if bit 14 is enabled, which digital input? (0-7)
 *   bits 4-3: when interested, interrupt every ...
 *     0 - ... 1 sample ...
 *     1 - ... 4 samples ...
 *     2 - ... 16 samples ...
 *     3 - ... 64 samples ...
 *   bit 2: ... and continue interrupting at this rate ...
 *     0 - ... until above interest criterias no longer met
 *     1 - ... forever (requires reset via write to bit 13, reg 0x2)
 *   bits 1-0: When not interested, interrupt ...
 *     0 - ... when buffer is half full
 *     1 - ... every 1ms
 *     2 - ... every 2ms
 *     3 - ... every 10ms
 *    
 * base + 0xa: reg used to detect core, reads alternate between 0x7558 & 0x1978
 * base + 0xc: PWM output controller (see info in wb_pwm.v, sample period 
 *  trigger is routed as external trigger #0)
 * base + 0xe: current state of all inputs (RO)
 *
 * * When only one ADC channel is enabled, the inter-sample gap (usually 40us)
 *   is skipped.  This allows much higher frequency sampling since the analog
 *   mux is never switched.
 *
 *
 * Input/Output Channels (TS-7558):
 * ================================
 *
 * If the upper 8 channels are not used for ADC (use the not_adc_chan module
 * parameter) they expose additional functionality using the digital inputs. 
 * The TS-7558 input channel map is:
 *
 *   0: ADC channel #0 (3.3V * 43.2kohm/(43.2kohm + 47kohm))
 *   1: ADC channel #1 (Voltage in * 43.2kohm/(43.2kohm + ?kohm))
 *   2: ADC channel #2 (ground)
 *   3: ADC channel #3 (5V / 2) 
 *   4: ADC channel #4 (external ADC channel#0)
 *   5: ADC channel #5 (external ADC channel#1)
 *   6: ADC channel #6 (external ADC channel#2)
 *   7: ADC channel #7 (external ADC channel#3)
 *   8: Quadrature counter delta #0 (input#4 as A, input#5 as B) 
 *   9: Quadrature counter delta #1 (input#6 as A, input#7 as B) 
 *  10: Current state of all inputs
 *  11: Counter delta #0 (counts posedges on input#0)
 *  12: Counter delta #1 (counts posedges on input#1)
 *  13: Counter delta #2 (counts posedges on input#2)
 *  14: Counter delta #3 (counts posedges on input#3)
 *  15: Glitch reg or timestamp
 *
 * * Channel 15 normally sets any bit as a one if it detects, during the course
 *   of the immediately preceeding sample period, any momentary change of a 
 *   digital input.  However, if the external trigger is enabled, this returns
 *   the latched value of a 12-bit up-counter that counts up at the rate of the 
 *   sample period as programmed in register base + 0x4
 *
 * Output (PWM) channels on the TS-7558:
 * 
 *   0: external output #0
 *   1: external output #1
 *   2: external output #2
 *   3: external output #3
 *   4: pin 23 on internal 26-pin .1" header labeled JTAG
 *   5: pin 25 on internal 26-pin .1" header labeled JTAG
 *   6: pin 21 on internal 26-pin .1" header labeled JTAG
 */
 
`undef TESTBENCH
`ifdef TESTBENCH
`timescale 1 ns / 1 ps
`endif
 
module adc7558(
  wb_clk_i,
  wb_rst_i,

  wbs_cyc_i,
  wbs_stb_i,
  wbs_we_i,
  wbs_adr_i,
  wbs_dat_i,
  wbs_sel_i,
  wbs_dat_o,
  wbs_ack_o,
  
  wbm_cyc_o,
  wbm_stb_o,
  wbm_adr_o,
  wbm_dat_o,
  wbm_ack_i,
  wbm_we_o,
  wbm_sel_o,
  
  adc_mux_o,
  adc_gain_2x_o,
  adc_cloop_en_o,
  adc_clk_o,
  adc_csn_o,
  adc_data_i,
  adc_enabled_o, /* 1-bit output for (e.g.) hijacking req'd ADC pins */
  sample_o, /* A single wb_clk high in synch. w/start of sampling */
  irq_o, 
  inputs_i,
  outputs_o, /* optional PWM outputs */
  
  clk_100mhz_i /* Constant 100Mhz to keep sw from having to know wb_clk */
);

/* Set to one for every channel thats not an ADC, any of the upper ones
 * not used take on special functionality the inputs_i port such as
 * quadrature, counters, and timing */
parameter not_adc_chan = 16'b1111111100000000;
 
localparam wb_freq = 75000000; /* Freq of wb_clk_i */
localparam inter_sample_gap_ns = 16000; /* time for analog mux to switch */
localparam spi_freq = 4000000; /* max freq of SPI (NOT TO EXCEED) */

integer delay1 = (inter_sample_gap_ns * 1e-9) / (1.0 / wb_freq) - 2.0;
integer delay2 = (1.0 / spi_freq) / (1.0 / wb_freq) / 2.0 - 1.5;

/* These parameters are passed to the PWM core, read wb_pwm.v for more info */
parameter pwm_nchan = 7;
parameter pwm_single_freq = 0; 
parameter pwm_use_exttrig0 = 1;
parameter pwm_use_exttrig1 = 0;
parameter pwm_hardcoded_freq0 = 0;
parameter pwm_hardcoded_freq1 = 0;

input wb_clk_i, wb_rst_i, wbs_cyc_i, wbs_stb_i, wbs_we_i;
input [31:0] wbs_adr_i;
input [15:0] wbs_dat_i;
input [1:0] wbs_sel_i;
output [15:0] wbs_dat_o;
output wbs_ack_o;

output wbm_cyc_o, wbm_stb_o, wbm_we_o;
output [31:0] wbm_adr_o;
output [15:0] wbm_dat_o;
output [1:0] wbm_sel_o;
input wbm_ack_i;

output [3:0] adc_mux_o;
output [3:0] adc_cloop_en_o;
output adc_gain_2x_o;
output adc_clk_o, adc_csn_o, adc_enabled_o;
input adc_data_i;
input [11:0] inputs_i;
output sample_o;
output irq_o;
output [7:0] outputs_o;

input clk_100mhz_i;

reg [15:0] wbm_dat;
assign wbm_dat_o = wbm_dat;
assign wbm_sel_o = 2'b11;
assign wbm_we_o = 1'b1;

reg [3:0] adc_mux;
reg adc_clk, adc_csn, adc_enable;
assign adc_clk_o = adc_clk;
assign adc_csn_o = adc_csn;
assign adc_mux_o = adc_mux;

reg [15:0] adc_period, wb_dat;
reg [12:0] wbm_adr;
reg wbm_cyc, wb_ack, irq;
assign wbm_adr_o = {19'd0, wbm_adr[12:0]};
assign wbm_cyc_o = wbm_cyc;
assign wbm_stb_o = wbm_cyc;
assign irq_o = irq & adc_enable;
assign wbs_dat_o = wb_dat;
assign wbs_ack_o = wb_ack;
reg [15:0] gain_en;
reg [15:0] chan_en;
reg [7:0] state;
reg [11:0] adc_dat;
reg [1:0] adc_trigclk_q;
reg adc_trigclk, adc_trig;
reg [3:0] chan;
reg [3:0] adc_clkcnt;
reg fsm_delay;
reg [15:0] fsm_dlycnt;
reg sample;
reg magicseq;
reg [14:0] nsamples, irq_thresh, nentries;
reg overflow;
reg [3:0] adc_cloop_en;

reg [11:0] counter_0, counter_1, counter_2, counter_3;
reg [11:0] quad_0, quad_1, inputs_q, delta;
reg i_anychg, i_counterchg, i_quadchg, i_quaddirchg, i_onechg;
reg [1:0] i_onechg_how;
reg [3:0] i_onechg_which;
reg [1:0] i_rate;
reg i_continue;
reg [1:0] i_notrate;
wire [15:0] interest_crit = {i_anychg, i_counterchg, i_quadchg, i_quaddirchg,
  i_onechg, i_onechg_how, i_onechg_which, i_rate, i_continue, i_notrate};

assign adc_cloop_en_o = adc_cloop_en;
assign sample_o = sample;
assign adc_enabled_o = adc_enable || state != 0 || wbm_cyc || fsm_delay;
reg adc_gain_2x, interested, interested_q, stay_interested, interesting;
assign adc_gain_2x_o = adc_gain_2x;
reg irq_thresh_recalc, div_start, end_of_sample;
reg [11:0] inputs_qq, inputs_qqq;
reg [15:0] div_num, div_denom;
reg [1:0] irq_thresh_calcstate;
wire div_done;
wire [15:0] div_q;
reg [11:0] trig_counter;
reg [3:0] ext_trig;
reg [1:0] ext_trig_input;
wire [15:0] pwmcore_dat;
reg pwm_trig, trip_irq;
wire pwmcore_ack;
wire pwm_en = (wbs_adr_i[3:1] == 3'd6);
wb_pwm #(
  .nchan(pwm_nchan),
  .single_freq(pwm_single_freq),
  .use_exttrig0(pwm_use_exttrig0),
  .use_exttrig1(pwm_use_exttrig1),
  .hardcoded_freq0(pwm_hardcoded_freq0),
  .hardcoded_freq1(pwm_hardcoded_freq1),
  .wb_freq(wb_freq)
) pwmcore(
  .wb_clk_i(wb_clk_i),
  .wb_rst_i(wb_rst_i),
  .wb_cyc_i(wbs_cyc_i),
  .wb_stb_i(wbs_stb_i && pwm_en && !wb_ack),
  .wb_dat_i(wbs_dat_i),
  .wb_dat_o(pwmcore_dat),
  .wb_ack_o(pwmcore_ack),
  .wb_we_i(wbs_we_i),

  .clk_100mhz_i(clk_100mhz_i),
  .exttrig0_i(pwm_trig),
  .exttrig1_i(pwm_trig),
  .pwm_o(outputs_o)
);
always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    ext_trig <= 4'd0;
    ext_trig_input <= 2'b00;
    i_anychg <= 1'b0;
    i_counterchg <= 1'b0;
    i_quadchg <= 1'b0;
    i_quaddirchg <= 1'b0;
    i_onechg <= 1'b0;
    i_onechg_how <= 2'b0;
    i_onechg_which <= 4'd0;
    i_rate <= 2'd0;
    i_continue <= 1'b0;
    i_notrate <= 2'd0;
    interested <= 1'b0;
    interested_q <= 1'b0;
    stay_interested <= 1'b0;
    adc_trigclk_q <= 2'b00;
    adc_trig <= 1'b0;
    trig_counter <= 12'd0;
    state <= 8'd0;
    adc_clk <= 1'b1;
    adc_csn <= 1'b1;
    adc_enable <= 1'b0;
    adc_period <= 16'd0;
    wbm_adr <= 13'd0;
    wbm_cyc <= 1'b0;
    wb_dat <= 16'd0;
    wb_ack <= 1'b0;
    gain_en <= 16'd0;
    chan_en <= 16'd0;
    chan <= 4'd0;
    adc_mux <= 4'd0;
    fsm_delay <= 1'b0;
    fsm_dlycnt <= 16'd0;
    adc_dat <= 12'd0;
    adc_clkcnt <= 4'd0;
    wbm_dat <= 16'd0;
    adc_cloop_en <= 4'd0;
    sample <= 1'b0;
    magicseq <= 1'b0;
    irq <= 1'b0;
    end_of_sample <= 1'b0;
    nsamples <= 15'd0;
    nentries <= 15'd0;
    irq_thresh <= 15'd2048;
    overflow <= 1'b0;
    adc_gain_2x <= 1'b0;
    div_start <= 1'b0;
    div_num <= 16'd0;
    div_denom <= 16'd0;
    irq_thresh_calcstate <= 2'd0;
    irq_thresh_recalc <= 1'b0;
    pwm_trig <= 1'b0;
    trip_irq <= 1'b0;
  end else begin
    ext_trig_input <= {ext_trig_input[0], inputs_i[ext_trig[2:0]]};
    wb_ack <= 1'b0;
    if (wbs_cyc_i && wbs_stb_i && !wb_ack) begin
      case (wbs_adr_i[3:1]) 
      3'd0: wb_dat <= chan_en;
      3'd1: begin
        nsamples <= 15'h0;
        nentries <= 15'h0;
        irq <= 1'b0;
        trip_irq <= 1'b0;
        if (!adc_enable && !wbm_cyc && state == 8'd0 && !fsm_delay)
          wb_dat[15] <= 1'b0;
        else wb_dat[15] <= 1'b1;
        wb_dat[14:0] <= {overflow, interested, 1'b0, wbm_adr[12:1]};
        overflow <= 1'b0;
      end
      3'd2: wb_dat <= adc_period;
      3'd3: wb_dat <= {ext_trig, adc_cloop_en, gain_en[7:0]};
      3'd4: wb_dat <= interest_crit;
      3'd5: wb_dat <= magicseq ? 16'h7558 : 16'h1978;
      3'd6: wb_dat <= pwmcore_dat;
      3'd7: wb_dat <= {4'd0, inputs_qq};
      endcase
      magicseq <= !magicseq;
      
      if (wbs_we_i) begin
        case (wbs_adr_i[3:1])
        3'd0: begin
          chan_en <= wbs_dat_i[15:0];
        end
        3'd1: begin
          adc_enable <= wbs_dat_i[15];
          if (!adc_enable && wbs_dat_i[15]) adc_trig <= 1'b1;
          interested <= wbs_dat_i[13];
          if (wbs_dat_i[13]) i_continue <= 1'b1;
        end
        3'd2: begin
          adc_period <= wbs_dat_i;
          irq_thresh_recalc <= 1'b1;
        end
        3'd3: begin
          gain_en[7:0] <= wbs_dat_i[7:0];
          adc_cloop_en <= wbs_dat_i[11:8];
          ext_trig <= wbs_dat_i[15:12];
        end
        3'd4: begin
          irq_thresh_recalc <= 1'b1;
          i_anychg <= wbs_dat_i[15];
          i_counterchg <= wbs_dat_i[14];
          i_quadchg <= wbs_dat_i[13];
          i_quaddirchg <= wbs_dat_i[12];
          i_onechg <= wbs_dat_i[11];
          i_onechg_how <= wbs_dat_i[10:9];
          i_onechg_which <= wbs_dat_i[8:5];
          i_rate <= wbs_dat_i[4:3];
          i_continue <= wbs_dat_i[2];
          i_notrate <= wbs_dat_i[1:0];
        end
        endcase
      end
      
      if (pwm_en) wb_ack <= pwmcore_ack;
      else wb_ack <= 1'b1;
    end
    
    pwm_trig <= 1'b0;
    adc_trigclk_q <= {adc_trigclk_q[0], adc_trigclk};
    if (adc_trigclk_q[0] != adc_trigclk_q[1]) begin
      if (!ext_trig[3]) begin
        adc_trig <= 1'b1;
        pwm_trig <= 1'b1;
      end
      trig_counter <= trig_counter + 1'b1;
    end
    if (ext_trig[3] && ext_trig_input == 2'b01) begin
      adc_trig <= 1'b1;
      pwm_trig <= 1'b1;
    end
    
    if (fsm_delay) begin
      if (fsm_dlycnt == 16'd0) fsm_delay <= 1'b0;
      fsm_dlycnt <= fsm_dlycnt - 1'b1;
    end else case (state)
      8'd0: if (adc_trig && adc_enable) begin
          adc_trig <= 1'b0;
          state <= 8'd1;
          sample <= 1'b1;
      end
      8'd1: begin
        if (chan_en[chan] && not_adc_chan[chan]) begin
          if (!wbm_cyc) state <= 8'd5;
        end else if (chan_en[chan]) begin
          state <= 8'd2;
          adc_mux <= chan;
          adc_gain_2x <= gain_en[chan[2:0]];
          /* optimization to skip inter-sample-gap if only sampling 1 chan */
          if (adc_mux != chan) begin
            fsm_delay <= 1'b1;
            fsm_dlycnt <= delay1; /* 40us to allow time for mux */
          end
        end else begin
          if (chan == 4'd15) begin
            state <= 8'd0;
            sample <= 1'b0;
            end_of_sample <= 1'b1;
          end
          chan <= chan + 1'b1;          
        end
      end
      8'd2: begin
        adc_csn <= 1'b0;
        fsm_delay <= 1'b1;
        fsm_dlycnt <= 16'd2; /* 20ns for CS# assert before CLK negedge */
        state <= 8'd3;
      end
      8'd3: begin
        adc_clk <= 1'b0;
        fsm_delay <= 1'b1;
        fsm_dlycnt <= delay2; /* CLK half-period */
        state <= 8'd4;
      end
      8'd4: begin /* rising ADC clk, latch/shift data */
        adc_clk <= 1'b1;
        fsm_delay <= 1'b1;
        fsm_dlycnt <= delay2;
        state <= 8'd3;
        adc_dat <= {adc_dat[10:0], adc_data_i};
        adc_clkcnt <= adc_clkcnt + 1'b1;
        if (adc_clkcnt == 4'd15) begin 
          adc_csn <= 1'b1;
          state <= 8'd1;
          wbm_dat <= {chan, adc_dat[11:0]};
          wbm_cyc <= 1'b1;
          chan <= chan + 1'b1;
          if (chan == 4'd15) begin
            state <= 8'd0;
            sample <= 1'b0;
            end_of_sample <= 1'b1;
          end
        end
      end
      8'd5: begin /* non-ADC sample */
        chan <= chan + 1'b1;
        wbm_cyc <= 1'b1;
        case (chan)
        4'd8: wbm_dat <= {chan, quad_0};
        4'd9: wbm_dat <= {chan, quad_1};
        4'd10: wbm_dat <= {chan, inputs_q};
        4'd11: wbm_dat <= {chan, counter_0};
        4'd12: wbm_dat <= {chan, counter_1};
        4'd13: wbm_dat <= {chan, counter_2};
        4'd14: wbm_dat <= {chan, counter_3};
        4'd15: if (!ext_trig[3]) wbm_dat <= {chan, delta};
          else wbm_dat <= {chan, trig_counter};
        endcase
        state <= 8'd1;
        if (chan == 4'd15) begin
          state <= 8'd0;
          sample <= 1'b0;
          end_of_sample <= 1'b1;
        end
      end
    endcase
      
    if (wbm_cyc && wbm_ack_i) begin
      wbm_cyc <= 1'b0;
      wbm_adr <= wbm_adr + 2'd2;
      nentries <= nentries + 1'b1;
      if (nentries[12:0] >= 13'd4096) overflow <= 1'b1;
    end
    if (end_of_sample && !wbm_cyc) begin /* Last sample in group? */
      end_of_sample <= 1'b0;
      nsamples <= nsamples + 1'b1;
      trip_irq <= 1'b0;
      if (trip_irq) irq <= 1'b1;
      if (!interested) begin
        if (nsamples >= irq_thresh) irq <= 1'b1;
        if (i_notrate == 2'd0 && nentries[11:0] >= 12'd3072) irq <= 1'b1;
      end
      case (i_rate)
      2'd0: begin
        if (interested) irq <= 1'b1;
        if (!i_continue && !stay_interested) interested <= 1'b0;
        stay_interested <= 1'b0;
      end
      2'd1: if (nsamples[1:0] == 2'd3) begin
        if (interested) irq <= 1'b1;
        if (!i_continue && !stay_interested) interested <= 1'b0;
        stay_interested <= 1'b0;
      end
      2'd2: if (nsamples[3:0] == 4'hf) begin
        if (interested) irq <= 1'b1;
        if (!i_continue && !stay_interested) interested <= 1'b0;
        stay_interested <= 1'b0;
      end
      2'd3: if (nsamples[5:0] == 6'h3f) begin
        if (interested) irq <= 1'b1;
        if (!i_continue && !stay_interested) interested <= 1'b0;
        stay_interested <= 1'b0;
      end
      endcase
    end

    if (interesting) begin
      interested <= 1'b1;
      if (!interested) trip_irq <= 1'b1;
      else stay_interested <= 1'b1;
    end

    div_start <= 1'b0;
    case (irq_thresh_calcstate)
    2'd0: if (irq_thresh_recalc) begin
      div_denom <= adc_period[14:0] + 1;
      irq_thresh_recalc <= 1'b0;
      case (i_notrate)
      2'd0: irq_thresh <= 15'h7fff;
      2'd1: if (adc_period[15]) irq_thresh <= 15'd0; else begin
        div_num <= 16'd1000;
        div_start <= 1'b1;
        irq_thresh_calcstate <= 2'd1; /* Wait for divider done */
      end
      2'd2: begin
        if (adc_period[15]) div_num <= 16'd2; else div_num <= 16'd2000;
        div_start <= 1'b1;
        irq_thresh_calcstate <= 2'd1; /* Wait for divider done */
      end
      2'd3: begin
        if (adc_period[15]) div_num <= 16'd10; else div_num <= 16'd10000;
        div_start <= 1'b1;
        irq_thresh_calcstate <= 2'd1; /* Wait for divider done */
      end
      endcase
    end
    2'd1: if (div_done) begin
      irq_thresh <= div_q[14:0] - 1'b1;
      $display("irq thresh=%d", div_q[14:0]);
      irq_thresh_calcstate <= 2'd0;
    end
    endcase
  end
end

/* divides div_num / div_denom and properly rounds to nearest int */
wire [15:0] div_qtmp;
assign div_q = div_qtmp[15:1] + div_qtmp[0];
divider irq_thresh_divider(
  .clk_i(wb_clk_i),
  .rst_i(wb_rst_i),
  .n_i({div_num[14:0], 1'b0}),
  .d_i(div_denom),
  .start_i(div_start),
  .done_o(div_done),
  .q_o(div_qtmp)
);

reg sample_q, counterchg, quadchg, quaddirchg, onechg;
reg [15:0] d, c0, c1, c2, c3, q0, q1;
always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    sample_q <= 1'b0;
    counter_0 <= 12'd0;
    counter_1 <= 12'd0;
    counter_2 <= 12'd0;
    counter_3 <= 12'd0;
    quad_0 <= 12'd0;
    quad_1 <= 12'd0;
    inputs_q <= 12'd0;
    inputs_qq <= 12'd0;
    inputs_qqq <= 12'd0;
    c0 <= 16'd0;
    c1 <= 16'd0;
    c2 <= 16'd0;
    c3 <= 16'd0;
    q0 <= 16'd0;
    q1 <= 16'd0;
    delta <= 12'd0;
    d <= 12'd0;
    interesting <= 1'b0;
    counterchg <= 1'b0;
    quadchg <= 1'b0;
    quaddirchg <= 1'b0;
    onechg <= 1'b0;
  end else begin
    sample_q <= sample;
    interesting <= 1'b0;
    if (!sample && sample_q) begin /* negedge sample */
      if (quadchg && i_quadchg) interesting <= 1'b1;
      if (quaddirchg && i_quaddirchg) interesting <= 1'b1;
      if (onechg && i_onechg) interesting <= 1'b1;
      if (counterchg && i_counterchg) interesting <= 1'b1;
      quadchg <= 1'b0;
      quaddirchg <= 1'b0;
      onechg <= 1'b0;
      counterchg <= 1'b0;
    end

    if (sample && !sample_q) begin /* posedge sample */
      inputs_q <= inputs_qqq;
      counter_0 <= c0[11:0];
      counter_1 <= c1[11:0];
      counter_2 <= c2[11:0];
      counter_3 <= c3[11:0];
      quad_0 <= q0[11:0];
      quad_1 <= q1[11:0];
      if (i_onechg_how == 2'd0 && 
        inputs_q[i_onechg_which] != inputs_qqq[i_onechg_which])
        onechg <= 1'b1;
      if (chan_en[8]) begin
        if (q0 != 16'd0) quadchg <= 1'b1;
        if (q0[11] != quad_0[11]) quaddirchg <= 1'b1;
      end
      if (chan_en[9]) begin
        if (q1 != 16'd0) quadchg <= 1'b1;
        if (q1[11] != quad_1[11]) quaddirchg <= 1'b1;
      end
      c0 <= 16'd0;
      c1 <= 16'd0;
      c2 <= 16'd0;
      c3 <= 16'd0;
      q0 <= 16'd0;
      q1 <= 16'd0;
      delta <= d;
      d <= 12'd0;
    end else begin
      inputs_qq <= inputs_i;
      inputs_qqq <= inputs_qq;
      d <= d | (inputs_qqq ^ inputs_qq);
      if ((inputs_qqq != inputs_qq) && i_anychg) interesting <= 1'b1;

      if (i_onechg_how == 2'd1 && i_onechg &&
        inputs_qqq[i_onechg_which] != inputs_qq[i_onechg_which])
        interesting <= 1'b1;

      if (i_onechg_how == 2'd2 && i_onechg &&
        inputs_qq[i_onechg_which]) interesting <= 1'b1;

      if (i_onechg_how == 2'd3 && i_onechg &&
        !inputs_qq[i_onechg_which]) interesting <= 1'b1;

      if (inputs_qq[0] && !inputs_qqq[0]) c0 <= c0 + 1'b1;
      if (inputs_qq[1] && !inputs_qqq[1]) c1 <= c1 + 1'b1;
      if (inputs_qq[2] && !inputs_qqq[2]) c2 <= c2 + 1'b1;
      if (inputs_qq[3] && !inputs_qqq[3]) c3 <= c3 + 1'b1;

      if ((inputs_qq[3:0] & ~inputs_qqq[3:0] & chan_en[14:11]) != 4'd0)
        counterchg <= 1'b1;

      if (inputs_qq[4] ^ inputs_qqq[4] ^ inputs_qq[5] ^ inputs_qqq[5]) begin
        if (inputs_qqq[4] ^ inputs_qq[5]) q0 <= q0 + 1'b1;
        else q0 <= q0 - 1'b1;
      end

      if (inputs_qq[6] ^ inputs_qqq[6] ^ inputs_qq[7] ^ inputs_qqq[7]) begin
        if (inputs_qqq[6] ^ inputs_qq[7]) q1 <= q1 + 1'b1;
        else q1 <= q1 - 1'b1;
      end
    end
  end
end

reg [14:0] adc_count;
reg [16:0] ts;
reg [15:0] adc_period_q;
reg adc_count_en;
reg [1:0] adc_enable_q;
always @(posedge clk_100mhz_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    adc_count <= 16'd0;
    adc_trigclk <= 1'b0;
    ts <= 17'd0;
    adc_count_en <= 1'b0;
    adc_enable_q <= 2'b00;
    adc_period_q <= 16'd0;
  end else begin
    adc_count_en <= 1'b0;
    adc_enable_q <= {adc_enable_q[0], adc_enable};
    adc_period_q <= adc_period;

    ts <= ts + 1'b1;
    if (ts == (adc_period_q[15] ? 17'd99999 : 17'd99)) begin    
      ts <= 17'd0;
      adc_count_en <= 1'b1;
    end
      
    if (adc_count_en) begin
      adc_count <= adc_count + 1'b1;
      if (adc_count == adc_period_q[14:0]) begin
        $display("%t ADC trig", $time);
        adc_trigclk <= !adc_trigclk;
        adc_count <= 16'd0;
      end
    end
    
    if (adc_enable_q[1] == 1'b0) begin
      ts <= 17'd0;
      adc_count <= 16'd0;
      adc_count_en <= 1'b0;
    end
    
  end
end
endmodule


module divider(
  n_i,
  d_i,
  q_o,
  clk_i,
  start_i,
  done_o,
  rst_i,
);

input [15:0] n_i;
input [15:0] d_i;
output [15:0] q_o;
input clk_i;
input rst_i;
input start_i;
output done_o;

reg [4:0] state;
reg [31:0] p;
reg [15:0] q;
assign q_o = q;
reg [16:0] x;
assign done_o = x[16];
wire [31:0] trial_subtract = {p[30:0], 1'b0} - {d_i, 16'd0};
always @(posedge clk_i or posedge rst_i) begin
  if (rst_i) begin
    p <= 32'd0;
    q <= 16'd0;
    x <= 16'd0;
  end else begin
    x <= {x[15:0], start_i};
    if (trial_subtract[31] == 1'b0) begin
      p <= trial_subtract;
      q <= {q[14:0], 1'b1};
    end else begin
      q <= {q[14:0], 1'b0};
      p <= {p[30:0], 1'b0};
    end
    if (start_i) p <= {16'd0, n_i};
  end
end

endmodule


`ifdef TESTBENCH
module test();

reg clk, rst, cyc, stb, we;
wire ack;
reg si;
reg [1:0] sel;
reg [15:0] idat, adr;
wire [15:0] odat, wbm_dat;
wire [3:0] adc_mux;
wire wbm_cyc;
wire [31:0] wbm_adr;
wire adc_csn, adc_gain_2x, adc_clk, irq;
reg adc_data;
reg [11:0] inputs;
adc7558 uut(
  .wb_clk_i(clk),
  .wb_rst_i(rst),

  .wbs_cyc_i(cyc),
  .wbs_stb_i(stb),
  .wbs_we_i(we),
  .wbs_adr_i({16'd0, adr}),
  .wbs_dat_i(idat),
  .wbs_sel_i(sel),
  .wbs_dat_o(odat),
  .wbs_ack_o(ack),

  .wbm_adr_o(wbm_adr),
  .wbm_dat_o(wbm_dat),
  .wbm_cyc_o(wbm_cyc),
  .wbm_ack_i(1'b1),
  
  .adc_mux_o(adc_mux),
  .adc_gain_2x_o(adc_gain_2x),
  .adc_clk_o(adc_clk),
  .adc_csn_o(adc_csn),
  .adc_data_i(adc_data),
  
  .clk_100mhz_i(clk),
  .irq_o(irq),
  .inputs_i(inputs)
);

reg [1:0] irq_q;
always @(posedge clk) begin
  irq_q <= {irq_q[0], irq};
  if (wbm_cyc) 
    $display("%t WBM write, adr %x, dat %x", $time, wbm_adr, wbm_dat);
  if (irq_q == 2'b01) $display("%t IRQ asserted", $time);
  else if (irq_q == 2'b10) $display("%t IRQ deasserted", $time);
  if (irq) wb_read16(16'd2);
end

reg [7:0] adc_count;
reg [15:0] adc_value, adc_sr;
always @(posedge adc_csn or posedge rst) begin
  if (rst) adc_value <= 16'd0;
  else adc_value <= adc_value + 2'd2;
end

always @(negedge adc_clk or posedge adc_csn) begin
  if (adc_csn) begin
    adc_count <= 8'd0;
    adc_data <= 1'bx;
  end else begin
    adc_count <= adc_count + 1'b1;
    adc_data <= adc_value[8'd15 - adc_count];
  end
end


task wb_read16;
  input [15:0] address_i;
  integer i;
  reg [7:0] clocks;
begin
  clocks = 8'd0;
  cyc = 1'b1;
  stb = 1'b1;
  we = 1'b0;
  adr = address_i;
  sel = 2'b11;
  idat = 16'hxxxx;
  @(posedge clk);
  while (!ack) begin
    @(posedge clk);
    clocks = clocks + 1'b1;
  end
  cyc = 1'b0; stb = 1'b0;
end
endtask

task wb_write16;
  input [15:0] address_i;
  input [15:0] data_i;
  integer i;
  reg [7:0] clocks;
begin
  clocks = 8'd0;
  cyc = 1'b1;
  stb = 1'b1;
  we = 1'b1;
  adr = address_i;
  sel = 2'b11;
  idat = data_i;
  @(posedge clk);
  while (!ack) begin
    @(posedge clk);
    clocks = clocks + 1'b1;
  end
  cyc = 1'b0; stb = 1'b0;
end
endtask

always #5 clk = !clk;
always #50000 inputs[0] = !inputs[0];

initial begin
  $dumpfile("dump.vcd");
  $dumpvars(2, test);
  $display("%d %d", uut.delay1, uut.delay2);
  rst = 1; clk = 0; inputs = 12'd0;
  cyc = 0; stb = 0; we = 0; adr = 0; sel = 0; idat = 0; 
  #100;
  rst = 0;
  #10;
  @(posedge clk);
  wb_write16(16'd0, 16'h800); /* Channel 11 enabled */
  wb_write16(16'd8, 16'h4008); /* irq */
  wb_write16(16'd4, 16'hc7); /* 200us sample period */
  wb_write16(16'd2, 16'h8000); /* Enable */
  wb_write16(16'hc, 16'h0c00); /* PWM 75% */
  #20000000;
  @(posedge clk);
  wb_read16(16'd0);

  #1000;
  $finish;
end

endmodule
`endif
