/* Copyright 2011, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS 
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, 
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, 
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL, 
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */

/* This core is for generating up to 8 PWM outputs.  Its features:
 *  - 12-bits of duty cycle (0 - 0%,  0xfff - 100%)
 *  - 2 external frequency trigger inputs
 *  - 2 internal frequency generators (each PWM chan picks one or the other)
 *  - Only one 16-bit write-only register conserves address decode
 *  - Various parameter "knobs" can save LUTs when features unneeded
 *  - PWM frequency can go as low as 10Hz
 *  - Small area (6 chans + 1 ext trig + 1 freq generator = 250LUTs)
 *
 * The core is controlled via opcodes written to its one write-only WISHBONE
 * register. There are actually 8 internal 13 bit registers that can be 
 * written but the basic format of the WISHBONE write is as follows:
 *
 *   bits 15-13: internal write-only register number
 *   bits 12-0: internal write-only register data
 * 
 * Internal write-only registers:
 *
 *   0: PWM control channel #0
 *     bit 15-13: must be 0
 *     bit 12: frequency select
 *     bit 11-0: duty cycle percent (0 - 0%, 0xfff - 100%)
 *   1: PWM control channel #1
 *     bit 15-13: must be 1
 *     bit 12: frequency select
 *     bit 11-0: duty cycle percent (0 - 0%, 0xfff - 100%)
 *   2: PWM control channel #2 ...
 *   3: PWM control channel #3 ...
 *   4: PWM control channel #4 ...
 *   5: PWM control channel #5 ...
 *   6: PWM frequency generator #0 (or PWM channel #6)
 *     bit 15-13: must be 6
 *     bit 12: period units (0 - 1 us, 1 - 100 us)
 *     bit 11-0: frequency period (multiplied by units)
 *   7: PWM frequency generator #1 (or PWM channel #7)
 *     bit 15-13: must be 7
 *     bit 12: period units (0 - 1 us, 1 - 100 us)
 *     bit 11-0: frequency period (multiplied by units)
 *
 * The frequency select (bit 12) of the PWM control channel #0-7 regs
 * selects which frequency to use and it depends on the parameters used 
 * during synthesis.  For instance, for bit == 0, precedence is interpreted
 * as follows:
 *   ... if hardcoded_freq0 is non-zero, then use that frequency, else ... 
 *   ... if use_exttrig0 is set, then use module port exttrig0_i, else ...
 *   ... use internal frequency generator #0
 *
 * Note that it is not possible to use all 8 channels and use any 
 * of the internal frequency generators.  The design will synthesize, but not
 * make much sense.
 * 
 * Compile-time parameters:
 * ========================
 * 
 * nchan: number of channels (default 6)
 *
 * single_freq: set to 1 if its not neccessary to allow users to select
 * between one of two frequencies for each channel.  This hardcodes to use
 * only freq #0.
 *
 * use_exttrig0, use_exttrig1: set either to one to use the external
 * trigger input. This overrides using the internal frequency generator.
 *
 * hardcoded_freq0, hardcoded_freq1: set either to non-zero to use a
 * hardcoded frequency input.  Expects a frequency in Hz.  This overrides
 * using the internal frequency generator AND using the external trigger.
 *
 * wb_freq: must be frequency in Hz of wb_clk_i input. (default 75Mhz)
 *
 * The compile-time parameters can be read from software by reading the
 * write-only register:  
 *  
 *   bits 15-8: Magic value "0x75"
 *   bit 7: hardcoded frequency is set for freq#1
 *   bit 6: hardcoded frequency is set for freq#0
 *   bit 5: uses external trigger input#1 for freq#1
 *   bit 4: uses external trigger input#0 for freq#0
 *   bit 3: uses only a single frequency input
 *   bits 2-0: number of PWM channels compiled-in
 *
 */

`undef TESTBENCH
`ifdef TESTBENCH
`timescale 1 ns / 1 ps
`endif
module wb_pwm(
  wb_clk_i,
  wb_rst_i,

  wb_cyc_i,
  wb_stb_i,
  wb_adr_i,
  wb_dat_i,
  wb_dat_o,
  wb_we_i,
  wb_ack_o,

  pwm_o,
  exttrig0_i,
  exttrig1_i,
  clk_100mhz_i /* For internal frequency generator */
);
input wb_clk_i;
input wb_rst_i;
input wb_cyc_i;
input wb_stb_i;
input wb_we_i;
input [31:0] wb_adr_i;
input [15:0] wb_dat_i;
output [15:0] wb_dat_o;
output wb_ack_o;

output [7:0] pwm_o;
input exttrig0_i;
input exttrig1_i;
input clk_100mhz_i;

parameter nchan = 6;
parameter single_freq = 0;
parameter use_exttrig0 = 1; 
parameter use_exttrig1 = 0; 
parameter hardcoded_freq0 = 0;
parameter hardcoded_freq1 = 0;
parameter wb_freq = 75000000; /* Used if a hardcoded_freq is set */

integer hardcoded_period0 = (1.0 / hardcoded_freq0) / (1.0 / wb_freq) - 1.5;
integer hardcoded_period1 = (1.0 / hardcoded_freq1) / (1.0 / wb_freq) - 1.5;

assign wb_dat_o[5:0] = 
  {use_exttrig1[0], use_exttrig0[0], single_freq[0], nchan[2:0]};
assign wb_dat_o[6] = (hardcoded_freq0 > 0);
assign wb_dat_o[7] = (hardcoded_freq1 > 0);
assign wb_dat_o[15:8] = 8'h75;

/* in units of 10 ns, max representable period = 168ms = 5.96hz */
reg [23:0] period0, counter0, period1, counter1;
reg [12:0] regmem[0:7]; /* 0 - 0%, 12'hfff - 100% */
reg [35:0] multiplied;
reg wb_ack, trigsel2, trigsel3;
reg [2:0] seq, seq2, seq3;
reg [7:0] pwm, pwm_0pct, pwm_100pct, pwm_rst, counter0_pwms, counter1_pwms;
assign pwm_o = pwm & ~pwm_0pct | pwm_100pct;
assign wb_ack_o = wb_ack;
reg [12:0] trigctl0, trigctl1;
wire inttrig0, inttrig1;

wire trig0 = use_exttrig0[0] ? exttrig0_i : inttrig0;
wire trig1 = use_exttrig1[0] ? exttrig1_i : inttrig1;

integer i;
wire trigsel = single_freq[0] ? 1'b0 : regmem[seq][12];
reg [23:0] multa, multb;
always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    multa <= 24'd0;
    multb <= 24'd0;
    wb_ack <= 1'b0;
    multiplied <= 36'd0;
    period0 <= hardcoded_freq0 > 0 ? hardcoded_period0 : 24'hffffff;
    counter0 <= 24'd0;
    period1 <= hardcoded_freq1 > 0 ? hardcoded_period1 : 24'hffffff;
    counter1 <= 24'd0;
    pwm <= 8'd0;
    pwm_0pct <= 8'hff;
    pwm_100pct <= 8'd0;
    trigsel2 <= 1'b0;
    trigsel3 <= 1'b0;
    seq <= 3'd0;
    seq2 <= 3'd0;
    seq3 <= 3'd0;
    trigctl0 <= {1'b1, 12'd99}; /* default 100Hz */
    trigctl1 <= {1'b1, 12'd99}; 
    for (i = 0; i < 8; i = i + 1) regmem[i] <= 13'd0;
    counter0_pwms <= 8'd0;
    counter1_pwms <= 8'd0;
    pwm_rst <= 8'd0;
  end else begin
    counter0 <= counter0 + 1'b1;
    counter1 <= counter1 + 1'b1;
    wb_ack <= 1'b0;
    if (wb_cyc_i && wb_stb_i && !wb_ack) wb_ack <= 1'b1;
    if (wb_cyc_i && wb_stb_i && wb_we_i && !wb_ack) begin
      if (wb_dat_i[15:13] == 3'd7) trigctl1 <= wb_dat_i[12:0];
      if (wb_dat_i[15:13] == 3'd6) trigctl0 <= wb_dat_i[12:0];
      if (wb_dat_i[15:13] < nchan) begin
	regmem[wb_dat_i[15:13]] <= wb_dat_i[12:0];
        if (wb_dat_i[11:0] != 12'd0) pwm_0pct[wb_dat_i[15:13]] <= 1'b0;
        else pwm_0pct[wb_dat_i[15:13]] <= 1'b1;
        if (wb_dat_i[11:0] != 12'hfff) pwm_100pct[wb_dat_i[15:13]] <= 1'b0;
        else pwm_100pct[wb_dat_i[15:13]] <= 1'b1;
      end
    end 

    if (seq >= (nchan - 1'b1)) seq <= 3'd0;
    else seq <= seq + 1'b1;
  
    if (hardcoded_freq1 > 0) begin
      if (counter1 >= hardcoded_period1) begin
        counter1 <= 24'd0;
        pwm_rst <= pwm_rst | counter1_pwms;
      end
    end else if (trig1) begin
      counter1 <= 24'd0;
      period1 <= counter1 + 1'b1;
      pwm_rst <= pwm_rst | counter1_pwms;
    end

    if (hardcoded_freq0 > 0) begin
      if (counter0 >= hardcoded_period0) begin
        counter0 <= 24'd0;
        pwm_rst <= pwm_rst | counter0_pwms;
      end
    end else if (trig0) begin
      counter0 <= 24'd0;
      period0 <= counter0 + 1'b1;
      pwm_rst <= pwm_rst | counter0_pwms;
    end

    multa <= {12'd0, regmem[seq][11:0]};
    multb <= (trigsel ? period1 : period0);
    trigsel2 <= trigsel;
    seq2 <= seq;
  
    multiplied <= multa * multb;
    trigsel3 <= trigsel2;
    seq3 <= seq2;

    if (pwm_rst[seq3]) begin
      pwm_rst[seq3] <= 1'b0;
      pwm[seq3] <= 1'b1;
    end
  
    if (multiplied[35:12] <= (trigsel3 ? counter1 : counter0)) begin
      pwm[seq3] <= 1'b0;
      counter1_pwms[seq3] <= trigsel3;
      counter0_pwms[seq3] <= !trigsel3;
    end
  end
end

periodic_trig freqgen0(
  .rst_i(wb_rst_i),
  .clk_i(wb_clk_i),
  .clk_100mhz_i(clk_100mhz_i),
  .timescale_i(trigctl0[12]),
  .count_i({20'd0, trigctl0[11:0]}),
  .trig_o(inttrig0)
);

periodic_trig freqgen1(
  .rst_i(wb_rst_i),
  .clk_i(wb_clk_i),
  .clk_100mhz_i(clk_100mhz_i),
  .timescale_i(trigctl1[12]),
  .count_i({20'd0, trigctl1[11:0]}),
  .trig_o(inttrig1)
);

endmodule

module periodic_trig(
  rst_i,
  clk_100mhz_i,
  timescale_i, /* 0 - 1 microseconds, 1 - 100 microseconds */
  count_i,

  clk_i,
  trig_o
);

input rst_i;
input clk_100mhz_i;
input timescale_i;
input [31:0] count_i;
input clk_i;
output trig_o;

parameter nbits = 12;

localparam N = nbits - 1;

reg [13:0] ts;
reg scale;
reg q;
always @(posedge clk_100mhz_i or posedge rst_i) begin
  if (rst_i) begin
    ts <= 14'd0;
    scale <= 1'b0;
    q <= 1'b0;
  end else begin
    scale <= timescale_i; /* ...so it can be from other clock domain */

    ts <= ts + 1'b1;
    if (ts == (scale ? 14'd9999 : 14'd99)) begin
      ts <= 14'd0;
      q <= !q;
    end
  end
end

reg [1:0] sync;
reg [31:0] timer;
reg trig;
assign trig_o = trig;
always @(posedge clk_i or posedge rst_i) begin
  if (rst_i) begin
    sync <= 2'd0;
    timer <= 32'd0;
    trig <= 1'b0;
  end else begin
    trig <= 1'b0;
    sync <= {sync[0], q};
    if (sync[0] != sync[1]) begin
      if (timer[N:0] >= count_i[N:0]) begin
        trig <= 1'b1;
        timer <= 32'd0;
      end else timer <= timer + 1'b1;
    end
  end
end

endmodule
  

`ifdef TESTBENCH
module test();

reg clk, rst, cyc, stb, trig0, trig1;
reg clk_100mhz;
reg [15:0] dat;
wire ack;
wire [7:0] pwm;
wb_pwm uut(
  .wb_clk_i(clk),
  .wb_rst_i(rst),
  .wb_cyc_i(cyc),
  .wb_stb_i(stb),
  .wb_dat_i(dat),
  .wb_we_i(1'b1),
  .wb_ack_o(ack),

  .exttrig0_i(trig0),
  .exttrig1_i(trig1),
  .pwm_o(pwm),
  .clk_100mhz_i(clk_100mhz)
);

task wb_write16;
  input [15:0] data_i;
  integer i;
  reg [7:0] clocks;
begin
  clocks = 8'd0;
  cyc = 1'b1;
  stb = 1'b1;
  dat = data_i;
  @(posedge clk);
  while (!ack) begin
    @(posedge clk);
    clocks = clocks + 1'b1;
  end
  cyc = 1'b0; stb = 1'b0;
  $display("clocks %d", clocks);
end
endtask

always #7 clk = !clk;
always #5 clk_100mhz = !clk_100mhz;

reg [7:0] counter = 8'd0;
always @(posedge clk) begin
  trig0 <= 1'b0;
  if (counter == 8'd99) begin
    trig0 <= 1'b1;
    counter <= 8'd0;
  end else counter <= counter + 1'b1;
end

initial begin
  $dumpfile("dump.vcd");
  $dumpvars(3, test);
  clk_100mhz = 0; clk = 0; rst = 1; cyc = 0; stb = 0; dat = 0;
  trig1 = 0;
  #100;
  rst = 0;
  wb_write16(16'h0800); /* Chan 0 = 50% */
  wb_write16(16'h2fff); /* Chan 1 = 100% */
  wb_write16(16'h4400); /* Chan 2 = 25% */
  #8800;
  wb_write16(16'h4c00); /* Chan 2 = 75% */
  #10000;
  $finish;
end
  
endmodule
`endif
